function [kijkl] = FEMmINT(img,msde,infs,it,i,isovalues)

% Copyright (c) 2023 Albert Argilaga
% For the publication: Generation of porous metamaterials with a 
% Physics-supervised deep learning model: FEM-GANs

addpath('.\iso2mesh-1.9.6\') %C:\Users\win11\Documents\MATLAB\DCGANmicro3D\iso2mesh-1.9.6
verbose=0; %
if infs == 0 && it==1 && i==1
    verbose=1; %
end

model = createpde('structural','static-solid');
[node,elem,~] = v2m(img,isovalues,msde,100,'cgalsurf'); %img, isovalues, opt, maxvol,method %external 3D image mesher %plotmesh(node,face)
%[node,elem,face]=vol2mesh(img,1:size(img,1),1:size(img,2),1:size(img,3),opt,maxvol,1,cgalsurf,isovalues);
%[node,elem,face,regions]=vol2mesh(img,ix,iy,iz,opt,maxvol,dofix,method,isovalues)
%
% convert a binary (or multi-valued) volume to tetrahedral mesh
%
% author: Qianqian Fang (q.fang at neu.edu)
%
% input:
%	 img: a volumetric binary image
%	 ix,iy,iz: subvolume selection indices in x,y,z directions
%	 opt: as defined in vol2surf.m
%	 maxvol: target maximum tetrahedral elem volume
%                when method='cgalmesh', maxvol can specify the target
%                for each label (subregion index) by the following syntax 
%                'label1=size1:label2=size2:...'
%	 dofix: 1: perform mesh validation&repair, 0: skip repairing
%	 method: 'cgalsurf' or omit: use CGAL surface mesher
%		 'simplify': use binsurface and then simplify
%		 'cgalmesh': use CGAL 3.5 3D mesher for direct mesh generation [new]
%
%		 generally speaking, 'cgalmesh' is the most robust path
%		 if you want to product meshes from binary or multi-region
%		 volumes, however, its limitations include 1) only accept 
%		 uint8 volume, and 2) can not extract meshes from gray-scale
%		 volumes. If ones goal is to process a gray-scale volume,
%		 he/she should use the 'cgalsurf' option. 'simplify' approach
%		 is not recommended unless other options has failed.
%	 isovalues: a list of isovalues where the levelset is defined
%
% output:
%	 node: output, node coordinates of the tetrahedral mesh
%	 elem: output, element list of the tetrahedral mesh, the last 
%	       column is the region ID
%	 face: output, mesh surface element list of the tetrahedral mesh
%	       the last column denotes the boundary ID
%        region: optional output. if opt.autoregion is set to 1, region
%              saves the interior points for each closed surface component
%
% -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)

while 1
    if size(elem,1)>0.8*128000 || size(node,1)>0.8*36000 ...
        || size(elem,1)<1.5*6500 || size(node,1)<1.5*1900 %im more it is just noise, break FEM
        disp('Exited FEM computation because elem or node too large or small') 
        kijkl=zeros(6,6)
        break
    end
    geometryFromMesh(model,node',elem(:,1:4)'); %back to MATLAB
    if model.Geometry.NumVertices>5000 || model.Geometry.NumVertices<250 %im more it is just noise, break FEM
        disp('Exited FEM computation because NumVertices>5000 or <250') 
        kijkl=zeros(6,6)
        break
    end

if verbose == 1
model.Geometry
figure(2)
pdeplot3D(model)
title('Mesh with Linear Tetrahedral Elements');
figure(3)
pdegplot(model,'FaceLabels','on')
view(30,30);
title('Boundary Value Problem geometry')
end

ratiofv= model.Geometry.NumVertices/model.Geometry.NumFaces;

structuralProperties(model,'YoungsModulus',9e10,'PoissonsRatio',0.17); %2.5e9  0.3 %70e9  0.22

vxyz = model.Geometry.vertexLabelLocations(); %cellLabelLocations(); faceLabelLocations(); edgeLabelLocations();
lbx=msde*0.3; ubx=size(img,1)-lbx; %depth to select vertices attached to BC %IDSall= find(vxyz{3,1} > 0);
lby=msde*0.3; uby=size(img,2)-lby;
lbz=msde*0.3; ubz=size(img,3)-lbz;
         
xd=vxyz{1,1}; %1 is x; 2 is y; 3 is z %Displacement unit
yd=vxyz{2,1}; %1 is x; 2 is y; 3 is z
zd=vxyz{3,1}; %1 is x; 2 is y; 3 is z
e(:,:,1)=xd*[1 0 0];
e(:,:,2)=yd*[0 1 0];
e(:,:,3)=zd*[0 0 1];
e(:,:,4)=0.5*(zd*[0 1 0]+yd*[0 0 1]);
e(:,:,5)=0.5*(zd*[1 0 0]+xd*[0 0 1]);
e(:,:,6)=0.5*(yd*[1 0 0]+xd*[0 1 0]); 
e=1e-5*e;     
IDSbc=find(vxyz{1,1} < lbx | vxyz{1,1} > ubx | vxyz{2,1} < lby | vxyz{2,1} > uby | vxyz{3,1} < lbz | vxyz{3,1} > ubz);
%IDSall=find(vxyz{1,1} > -100);
%ratbc=size(IDSbc,1)/size(IDSall,1);

if verbose == 1
dispstr = ['Parallel solving (x6), Albert Argilaga, 2022. Available threads: ',num2str(maxNumCompThreads)];
disp(dispstr)
end

    parfor i=1:6
    %IDSbc=bc6(vxyz,lbx,ubx,lby,uby,lbz,ubz,8); %input i in the last parameter for parallel loading
    for j=IDSbc'
        structuralBC(model,'vertex',j,'Displacement', e(j,1:3,i));
    end
    if verbose == 1
    tic;
    end
    result = solve(model);
%     if i == 1 %plot one solution
%         figure
%         %pdeplot3D(model,'ColorMapData',result.VonMisesStress)
%         %title('von Mises stress')
%         %pdeplot3D(model,'ColorMapData',result.Stress.sxx)
%         %title('Stress s_{11}')
%         pdeplot3D(model,'ColorMapData',result.Displacement.ux)
%         title('FEM solution - Dispalcements u_{11}')
%         colormap('jet')
%     end
    intr=0.5; %resolution of the homogenization interpolation
    [X,Y,Z] = meshgrid(linspace(0+1*lbx,size(img,3)-1*lbx,round(intr*(size(img,3))))); %Z = ones(size(X))*size(img,3)/2;
    intrpStress = interpolateStress(result,X,Y,Z);
    kijkl(:,i)=[sum(intrpStress.xx,'omitnan'); sum(intrpStress.yy,'omitnan'); ...
                sum(intrpStress.zz,'omitnan'); sum(intrpStress.yz,'omitnan'); ...
                sum(intrpStress.xz,'omitnan'); sum(intrpStress.xy,'omitnan')];
    if verbose == 1
    dispstr = ['FEM solution ',num2str(i),' of ',num2str(6),' Time taken: ',num2str(toc)];
    disp(dispstr)
    end
    end

intr=1;
kijkl=kijkl/1e-5/(intr*(size(img,3))-2*lbx)^3*ratiofv; %
%fprintf ('Homogenization and parallelization by Albert Argilaga, 2022 \n');
%(kijkl(1,1)+kijkl(2,2)+kijkl(3,3))/3;
kplot=kijkl/max(max(kijkl))*256;

if verbose == 1
kijkl
figure(4)
image(kplot)
axis square
%structuralBoundaryLoad (model,'face',IDSz0f,'SurfaceTraction',[0;0;+1e6]);figure
% result = solve(model);
% 
% figure
% pdeplot3D(model,'ColorMapData',result.Stress.sxx)
% title('Stress s_{11}')

figure(5)
pdeplot3D(model)
title('Mesh with Quadratic Tetrahedral Elements');
figure(1)
end

break
end

% function [IDSbc] = bc6(vxyz,lbx,ubx,lby,uby,lbz,ubz,i)
% switch i
%     case 1
%     IDSbc=find(vxyz{1,1} < lbx | vxyz{1,1} > ubx);
%     case 2          
%     IDSbc=find(vxyz{2,1} < lby | vxyz{2,1} > uby);
%     case 3
%     IDSbc=find(vxyz{3,1} < lbz | vxyz{3,1} > ubz);
%     case 4
%     IDSbc=find(vxyz{2,1} < lby | vxyz{2,1} > uby | vxyz{3,1} < lbz | vxyz{3,1} > ubz);
%     case 5
%     IDSbc=find(vxyz{1,1} < lbx | vxyz{1,1} > ubx | vxyz{3,1} < lbz | vxyz{3,1} > ubz);
%     case 6
%     IDSbc=find(vxyz{1,1} < lbx | vxyz{1,1} > ubx | vxyz{2,1} < lby | vxyz{2,1} > uby);
%     otherwise
%     disp('Rigid wall box loading (6 walls)')
%     IDSbc=find(vxyz{1,1} < lbx | vxyz{1,1} > ubx | vxyz{2,1} < lby | vxyz{2,1} > uby | vxyz{3,1} < lbz | vxyz{3,1} > ubz);
% end
% end

end